import 'dart:async';

import './src/core/gateway_client.dart';
import './src/core/peerconnection.dart';
import './src/utils/logger.dart';
import './src/utils/report.dart';
import './src/utils/statsfilter.dart';

import 'package:flutter_webrtc/webrtc.dart';
import './src/core/config.dart';

class AgoraStream {
  /// Marks whether this stream contains an audio track.
  bool audio;

  /// Marks whether this stream contains a video track.
  bool video;

  /// @nodoc
  bool screen;

  /// Marks whether this stream in local.
  bool isLocal = true;

  /// The stream width
  int width = 640;

  /// The stream height
  int height = 480;

  /// The stream frame rate
  int frameRate = 15;

  /// The stream bit rate
  int bitRate = 500;

  /// The stream ID
  int streamId;

  /// Sets using the front or rear camera
  String facingMode = "user";

  /// @nodoc
  String sid;

  /// @nodoc
  MediaStream mediaStream;

  /// @nodoc
  RTCVideoRenderer render;

  /// @nodoc
  Map<String, dynamic> mediaConstraints;

  /// @nodoc 
  GatewayClient gw;

  /// @nodoc
  PeerConnection pc;

  /// @nodoc
  Map<String, dynamic> get streamInfo {
    return {
      "audio": audio,
      "video": video,
      "screen": screen,
      "isLocal": isLocal,
      "width": width,
      "height": height,
      "frameRate": frameRate,
      "bitRate": bitRate,
      "streamId": streamId,
      "facingMode": facingMode
    };
  }

  /// Marks whether this stream mutes an audio track.
  bool get isAudioMuted {
    if (mediaStream != null && audio == true) {
      if (mediaStream.getAudioTracks().length <= 0) {
        Log.warn(
            '[$streamId]: Audio track is not exist when invoke [isAudioMuted]');
        return false;
      }
      var track = mediaStream.getAudioTracks()[0];
      return !track.enabled;
    } else {
      return false;
    }
  }

  /// Marks whether this stream mutes an video track.
  bool get isVideoMuted {
    if (mediaStream != null && (video == true || screen == true)) {
      if (mediaStream.getVideoTracks().length <= 0) {
        Log.warn(
            '[$streamId]: Video track is not exist when invoke [isVideoMuted]');
        return false;
      }
      var track = mediaStream.getVideoTracks()[0];
      return !track.enabled;
    } else {
      return false;
    }
  }

  /// Marks whether this stream have been subscribed
  bool get hasSubscribed {
    if (isLocal == false && mediaStream != null) {
      return true;
    }
    return false;
  }

  /// Creates a Stream Object
  /// 
  /// This method creates and returns a stream object. option contains follow property:
  /// - video (bool)
  /// - auido (bool)
  /// - facingMode ("environment" | "user") 
  static AgoraStream createStream(Map<String, dynamic> option) {
    return new AgoraStream(option);
  }

  /// Creates a Stream Object
  /// 
  /// This method creates and returns a stream object. option contains follow property:
  /// - video (bool)
  /// - auido (bool)
  /// - facingMode ("environment" | "user") 
  AgoraStream(Map<String, dynamic> option) {
    video = option["video"] == null ? false : option["video"];
    audio = option["audio"] == null ? false : option["audio"];
    screen = option["screen"] == null ? false : option["screen"];

    if (video == false && audio == false && screen == false) {
      Log.error('[$streamId]: Video, audio, screen should not all false');
      throw "Video, audio, screen should not all false";
    }

    facingMode =
        option["facingMode"] == null ? facingMode : option["facingMode"];
    isLocal = option["isLocal"] == null ? isLocal : option["isLocal"];
    streamId = option["streamId"];
  }

  /// Initializes the Stream Object
  /// 
  /// This method initializes the local stream object.
  Future init() async {
    Map<String, dynamic> contraint = {};

    if (audio == true) {
      // set attribute in the future
      contraint["audio"] = true;
    } else {
      contraint["audio"] = false;
    }

    if (video == true || screen == true) {
      contraint["video"] = {
        'mandatory': {
          'minWidth': width.toString(),
          'minHeight': height.toString(),
          'minFrameRate': frameRate.toString(),
        },
        'facingMode': facingMode,
        'optional': [],
      };
    } else {
      contraint["video"] = false;
    }

    mediaConstraints = contraint;
    var reportOption = new Map<String, dynamic>.from(contraint);
    EventReportManager.report(sid, "Stream.init", reportOption);

    if (screen == true) {
      try {
        mediaStream = await navigator.getDisplayMedia(contraint);
        return Future.value(this);
      } catch (err) {
        return Future.error(err);
      }
    }

    if (video == true) {
      try {
        mediaStream = await navigator.getUserMedia(contraint);
        return Future.value(this);
      } catch (err) {
        return Future.error(err);
      }
    }
  }

  /// Retrieves the Stream ID
  /// 
  /// This method retrieves the stream ID.
  int getId() {
    return streamId;
  }

  /// Sets the Video Bit Rate
  void setVideoBitRate(int bitRate) {
    this.bitRate = bitRate;
  }

  /// Sets the Video Frame Rate
  void setVideoFrameRate(int frameRate) {
    this.frameRate = frameRate;
  }

  /// Sets the Video Resolution
  void setVideoResolution(int width, int height) {
    this.width = width;
    this.height = height;
  }

  /// Sets the Video Profile
  /// 
  /// This method sets the video encoding profile for the local stream. Each video encoding profile includes a set of parameters, such as the resolution, frame rate, and bitrate.
  void setVideoProfile(String profileName) {
    var profile = AgoraProfile[profileName];
    if (profile == null) {
      return;
    }
    width = profile["width"];
    height = profile["height"];
    frameRate = profile["frameRate"];
    bitRate = profile["bitRate"];
  }

  /// @nodoc
  void setScreenProfile() {}

  /// Retrieves the Audio Flag
  bool hasAudio() {
    return !!audio;
  }

  /// Retrieves the Video Flag
  bool hasVideo() {
    return !!video;
  }

  /// @nodoc
  bool hasScreen() {
    return !!screen;
  }

  /// Disables the Audio
  /// 
  /// This method disables the audio track in the stream.
  void muteAudio() {
    Log.debug('[$streamId]: mute audio');
    var track = this.mediaStream.getAudioTracks()[0];
    if (track == null) {
      Log.warn('[$streamId]: Audio track is not exist when invoke [muteAudio]');
    } else {
      if (track.enabled != false) {
        track.enabled = false;
        gw?.muteAudio(this);
      }
    }
  }

  /// Disables the Video
  /// 
  /// This method disables the video track in the stream.
  void muteVideo() {
    Log.debug('[$streamId]: mute video');
    var track = this.mediaStream.getVideoTracks()[0];
    if (track == null) {
      Log.warn('[$streamId]: Video track is not exist when invoke [muteVideo]');
    } else {
      if (track.enabled != false) {
        track.enabled = false;
        gw?.muteVideo(this);
      }
    }
  }

  /// Enables the Audio
  /// 
  /// This method enables the audio track in the stream.
  void unmuteAudio() {
    Log.debug('[$streamId]: unmute audio');
    var track = this.mediaStream.getAudioTracks()[0];
    if (track == null) {
      Log.warn(
          '[$streamId]: Audio track is not exist when invoke [unmuteAudio]');
    } else {
      if (track.enabled != true) {
        track.enabled = true;
        gw?.unmuteAudio(this);
      }
    }
  }

  /// Enables the Video
  /// 
  /// This method enables the video track in the stream.
  void unmuteVideo() {
    Log.debug('[$streamId]: unmute video');
    var track = this.mediaStream.getVideoTracks()[0];
    if (track == null) {
      Log.warn(
          '[$streamId]: Video track is not exist when invoke [unmuteVideo]');
    } else {
      if (track.enabled != true) {
        track.enabled = true;
        gw?.unmuteVideo(this);
      }
    }
  }

  /// Switches the Media Input Device
  /// 
  /// This method switches between the media input devices:
  void switchDevice() async {
    if (screen == true) {
      Log.error('[$streamId]: Not support switch device when sharing screen');
      throw Exception("Not support switch device when sharing screen");
    }
    facingMode = facingMode == "user" ? "environment" : "user";
    var track = mediaStream.getVideoTracks()[0];
    track.switchCamera();
  }

  /// Plays the Audio/Video Stream
  /// 
  /// This method plays the video or audio stream.
  void play({String mode = "cover"}) async {
    Log.debug('[$streamId]: play');
    if (mediaStream == null) {
      Log.debug('[$streamId]: Can\'t play when mediaStream is null');
      throw "Can't play when mediaStream is null";
    }
    var render = new RTCVideoRenderer();
    render.initialize();
    if (mode == "cover") {
      render.objectFit = RTCVideoViewObjectFit.RTCVideoViewObjectFitCover;
    }
    if (mode == "contain") {
      render.objectFit = RTCVideoViewObjectFit.RTCVideoViewObjectFitContain;
    }
    // delay 100ms, or will throw error in native layer, may cause by unfinishing render.initialize
    // render.initialize dose not have finsih callback or message
    await Future.delayed(const Duration(milliseconds: 100));
    render.srcObject = mediaStream;
    this.render = render;
  }

  /// Returns Whether the Stream is Playing
  bool isPlaying() {
    // Log.debug('[$streamId]: isPlaying');
    return render != null;
  }

  /// Gets Connection Statistics
  /// 
  /// This method gets the connection statistics of the stream.
  Map<String, String> getStats() {
    if (pc == null) {
      return null;
    }
    if (pc.stats == null) {
      return null;
    }
    if (isLocal) {
      return StatsFilter.publishStatsFilter(pc.stats);
    } else {
      return StatsFilter.subscribeStatsFilter(pc.stats);
    }
  }

  /// Closes the Audio/Video Stream
  /// 
  /// This method closes the video/audio stream.
  void close() {
    Log.debug('[$streamId]: close');
    mediaStream?.dispose();
    mediaStream = null;
    render?.dispose();
    render = null;
    mediaConstraints = null;
    gw = null;
    pc = null;
  }

  /// Stops the Audio/Video Stream Playback
  /// 
  /// Call this method to stop playing the stream set by Stream.play.
  void stop() {
    Log.debug('[$streamId]: stop');
    render?.dispose();
    render = null;
  }
}
