import 'dart:async';
import 'dart:convert';
import '../utils/logger.dart';
import 'package:http/http.dart' as http;
import 'config.dart';
import 'error_code.dart';

class ChooseServer {
  String clientId;
  String sid;
  String appId;
  String cname;
  int uid;
  http.Client httpClient;
  Completer completer;

  ChooseServer({this.clientId, this.uid, this.sid, this.appId, this.cname}) {
    httpClient = new http.Client();
  }

  Future<Map<String, dynamic>> singleRequest(String domain) async {
    final String url = "https://$domain/api/v1?action=wrtc_gateway";
    try {
      final Map<String, dynamic> body = {
        'flag': 4096,
        'opid': 133,
        'cname': cname,
        'detail': {},
        'key': appId,
        'sid': sid,
        'ts': DateTime.now().microsecondsSinceEpoch ~/ 1000,
        'uid': uid,
      };

      final Map<String, String> headers = {
        'Content-type': 'application/json; charset=UTF-8',
        'Accept': 'application/json',
        'X-Packet-Service-Type': '0',
        'X-Packet-URI': '69'
      };

      final response =
          await httpClient.post(url, body: json.encode(body), headers: headers);

      // Error on HTTP code
      if (response.statusCode != 200) {
        final code = response.statusCode;
        return Future.error({
          "url": url,
          "body": body,
          "reason": "HTTP request failed, statusCode: $code",
        });
      }

      // Error on HTTP response body
      final String resbody = response.body;
      var res;

      try {
        res = jsonDecode(resbody);
        if (res["code"] == null) {
          throw Error();
        }
      } catch (err, stack) {
        return Future.error({
          "url": url,
          "body": body,
          "reason":
              "Request choose server failed with unexpected body $resbody, err: $err, stack: $stack",
        });
      }

      // Error on AP
      if (res["code"] != 0) {
        var err = APErrorCode[res["code"]];
        if (err == null) {
          err = res["code"];
        }
        return Future.error({
          "url": url,
          "body": body,
          "reason": "Get server node failed [$err]",
        });
      }

      // Error on addresses
      if (res["addresses"] == null || res["addresses"].length == 0) {
        return Future.error({
          "url": url,
          "body": body,
          "reason": "Get server node failed [EMPTY_ADDRESS_RESPONSE]",
        });
      }

      final addresses = res["addresses"];
      final cid = res["cid"];
      final List<Map<String, String>> addrs = [];
      addresses.forEach((address) {
        final Map<String, String> host = {
          "address":
              "${address["ip"].replaceAll(new RegExp(r'\.'), '-')}.edge.agora.io:${address["port"]}",
          "ticket": address["ticket"],
        };
        addrs.add(host);
      });
      return Future.value({
        "url": url,
        "body": {
          "apResponse": res,
          "addresses": addrs,
          "cid": cid,
        },
        "reason": ""
      });
    } catch (err, stack) {
      return Future.error({
        "url": url,
        "body": 'body',
        "reason": "AP request failed: $err, stack: $stack",
      });
    }
  }

  Future<Map<String, dynamic>> request() async {
    if (completer != null) {
      return Future.error("duplicate reuqest");
    }

    Completer c = Completer<Map<String, dynamic>>();
    completer = c;

    final Timer timeoutTimer = new Timer(Duration(seconds: 5), () {
      Log.debug("[$clientId]: choose server request timeout");
      c.completeError("choose server request timeout");
      completer = null;
    });

    final Timer backupTimer = new Timer(Duration(seconds: 2), () {
      WEBCS_DOMAIN_BACKUP_LIST.forEach((dmain) {
        singleRequest(dmain).then((Map<String, dynamic> res) {
          timeoutTimer?.cancel();
          _requestSuccess(res, c);
        }, onError: _requestFailed).catchError(_requestError);
      });
    });

    WEBCS_DOMAIN.forEach((dmain) {
      singleRequest(dmain).then((Map<String, dynamic> res) {
        timeoutTimer?.cancel();
        backupTimer?.cancel();
        _requestSuccess(res, c);
      }, onError: _requestFailed).catchError(_requestError);
    });
    return c.future;
  }

  _requestSuccess(Map<String, dynamic> res, Completer c) {
    // TODO choose report
    final String url = res["url"];
    final Map<String, dynamic> servers = res["body"];
    Log.debug(
        "[$clientId]: choose server success, url: $url, servers: $servers");
    c.complete(servers);
    completer = null;
  }

  _requestFailed(res, StackTrace stack) {
    // TODO choose report
    final url = res["url"];
    final reason = res["reason"];
    // Log.debug("[$clientId]: choose server failed, url: $url");
    // Log.debug("[$clientId]: reason: $reason");
  }

  _requestError(err, StackTrace stack) {
    // Log.debug("[$clientId]: choose server error, error: $err");
    // Log.debug("[$clientId]: stack: $stack");
  }

  cancel() {
    if (completer != null) {
      Log.debug("[$clientId]: choose server cancel");
      completer.complete(null);
    }
  }
}
