import 'dart:async';
import './error_code.dart';
import './exception.dart';

import './choose_server.dart';
import '../utils/logger.dart';

import './signal.dart';
import './message.dart';
import '../../agora_stream.dart';
import './peerconnection_controller.dart';

const String DISCONNECTED = "DISCONNECTED";
const String CONNECTING = "CONNECTING";
const String CONNECTED = "CONNECTED";
const String JOINED = "JOINED";
const String LEAVING = "LEAVING";
const String LEAVED = "LEAVED";

class GatewayClient {
  Signal signal;
  PeerConnectionController pcController;

  String mode;
  String codec;
  String appId;
  String token;
  String channel;
  String sid;
  String clientId;
  String role = "host";
  String _state = LEAVED;
  int uid;

  Function onConnectionStateChange;
  ChooseServer chooseServer;
  List<Map<String, String>> serverList;

  set state(String state) {
    var prvState = _state;
    _state = state;
    if (onConnectionStateChange != null) {
      onConnectionStateChange(prvState, state);
    }
  }

  get state {
    return _state;
  }

  get gwEvent {
    return signal?.gwEvent;
  }

  get _sessionInfo {
    return {
      "uid": uid,
      "app_id": appId,
      "session_id": sid,
      "process_id": Log.processId,
      "role": role,
      "mode": mode,
      "codec": codec,
      "channel_key": token,
      "channel_name": channel,
      // Fix
      "has_changed_gateway": false,
    };
  }

  GatewayClient({this.appId, this.codec, this.mode, this.clientId, this.sid});

  Future<Map<String, dynamic>> join(
      String token, String channel, int uid) async {
    Log.debug("[$clientId]: (gatewayClient) join with: $token, $channel, $uid");
    if (token == null) {
      this.token = appId;
    }
    this.channel = channel;
    this.uid = uid;

    chooseServer = new ChooseServer(
      clientId: clientId,
      uid: uid,
      sid: sid,
      appId: appId,
      cname: channel,
    );

    // get best gateway server address and ticket
    var csRes;
    try {
      csRes = await chooseServer.request();
    } catch (exception, stackTrace) {
      throw AgoraException(CONNECT_CHOOSE_SERVER_FALIED,
          "connect choose server failed", exception, stackTrace);
    }
    var apResponse = csRes["apResponse"];
    final cid = csRes["cid"];
    serverList = csRes["addresses"];

    final String url = serverList[0]["address"];
    final String ticket = serverList[0]["ticket"];
    apResponse["ticket"] = ticket;

    // new signal with gateway server address
    Log.debug("[$clientId]:(gatewayClient) Connect to $url");
    state = CONNECTING;
    try {
      signal = await Signal.connect(url, clientId);
    } catch (exception, stackTrace) {
      throw AgoraException(CONNECT_WEBSOCKET_FALIED, "connect websocket failed",
          exception, stackTrace);
    }
    signal.onClose = _onSiganlClose;
    state = CONNECTED;
    pcController = PeerConnectionController(signal, clientId, _sessionInfo);

    // join message
    final Map<String, dynamic> joinReq =
        Message.generate("join", [_sessionInfo, apResponse]);
    final Map<String, dynamic> joinRes = await signal.request(joinReq);
    if (joinRes["_result"] != "success") {
      final String reason = "${joinRes["error_code"]}: ${joinRes["error_str"]}";
      Log.error("[$clientId]:(gatewayClient) Join request failed: $reason");
      throw AgoraException(
          JOIN_REQUEST_FAILED, "Join request failed", "$reason", null);
    }

    // return uid
    uid = joinRes["_message"]["uid"];
    state = JOINED;
    return {
      "uid": uid,
      "cid": cid,
    };
  }

  Future<AgoraStream> publish(AgoraStream stream) async {
    Log.debug("[$clientId]:(gatewayClient) publish");
    try {
      await pcController.publish(stream);
      stream.gw = this;
      // sync mute state
      if (stream.isAudioMuted) {
        muteAudio(stream);
      } else {
        unmuteAudio(stream);
      }
      if (stream.isVideoMuted) {
        muteVideo(stream);
      } else {
        unmuteVideo(stream);
      }
    } catch (exception, stackTrace) {
      Log.error(
          "[$clientId]:(gatewayClient) Publish Failed: ${exception.toString()}}");
      throw AgoraException(
          PUBLISH_FAILED, "publish failed", exception, stackTrace);
    }
    return stream;
  }

  Future<AgoraStream> unpublish(AgoraStream stream) async {
    Log.debug("[$clientId]:(gatewayClient) unpublish");
    try {
      await pcController.unpublish(stream);
      stream.gw = null;
    } catch (exception, stackTrace) {
      Log.error(
          "[$clientId]:(gatewayClient) Unpublish Failed: ${exception.toString()}");
      throw AgoraException(
          UNPUBLISH_FAILED, "Unpublish failed", exception, stackTrace);
    }
    return stream;
  }

  Future<AgoraStream> subscribe(AgoraStream stream) async {
    Log.debug("[$clientId]:(gatewayClient) subscribe");
    try {
      await pcController.subscribe(stream);
    } catch (exception, stackTrace) {
      Log.error(
          "[$clientId]:(gatewayClient) Subscribe Failed: ${exception.toString()}");
      throw AgoraException(
          SUBSCRIBE_FAILED, "Subscribe failed", exception, stackTrace);
    }
    return stream;
  }

  Future<AgoraStream> unsubscribe(AgoraStream stream) async {
    Log.debug("[$clientId]:(gatewayClient) unsubscribe");
    try {
      await pcController.unsubscribe(stream);
    } catch (exception, stackTrace) {
      Log.error(
          "[$clientId]:(gatewayClient) Unsubscribe Failed: ${exception.toString()}");
      throw AgoraException(
          UNSUBSCRIBE_FAILED, "Unsubscribe failed", exception, stackTrace);
    }
    return stream;
  }

  muteVideo(AgoraStream stream) {
    var muteVideoReq;
    if (stream.isLocal) {
      muteVideoReq = Message.generate("muteVideo", [
        {'stream_id': stream.getId()}
      ]);
    } else {
      muteVideoReq = Message.generate("muteVideoRemote", [
        {'stream_id': stream.getId()}
      ]);
    }
    Log.debug(
        "[$clientId]:(gatewayClient) muteVideo with message: $muteVideoReq");
    signal.send(muteVideoReq);
  }

  muteAudio(AgoraStream stream) {
    var muteAudioReq;
    if (stream.isLocal) {
      muteAudioReq = Message.generate("muteAudio", [
        {'stream_id': stream.getId()}
      ]);
    } else {
      muteAudioReq = Message.generate("muteAudioRemote", [
        {'stream_id': stream.getId()}
      ]);
    }
    Log.debug(
        "[$clientId]:(gatewayClient) muteAudio with message: $muteAudioReq");
    signal.send(muteAudioReq);
  }

  unmuteVideo(AgoraStream stream) {
    var unmuteVideoReq;
    if (stream.isLocal) {
      unmuteVideoReq = Message.generate("unmuteVideo", [
        {'stream_id': stream.getId()}
      ]);
    } else {
      unmuteVideoReq = Message.generate("unmuteVideoRemote", [
        {'stream_id': stream.getId()}
      ]);
    }
    Log.debug(
        "[$clientId]:(gatewayClient) unmuteVideo with message: $unmuteVideoReq");
    signal.send(unmuteVideoReq);
  }

  unmuteAudio(AgoraStream stream) {
    var unmuteVideoReq;
    if (stream.isLocal) {
      unmuteVideoReq = Message.generate("unmuteAudio", [
        {'stream_id': stream.getId()}
      ]);
    } else {
      unmuteVideoReq = Message.generate("unmuteAudioRemote", [
        {'stream_id': stream.getId()}
      ]);
    }
    Log.debug(
        "[$clientId]:(gatewayClient) unmuteAudio with message: $unmuteVideoReq");
    signal.send(unmuteVideoReq);
  }

  closeSubConnection(int id) {
    Log.debug("[$clientId]:(gatewayClient) closeSubConnection id: $id");
    pcController?.closeSubConnection(id);
  }

  closePubConnection(int id) {
    Log.debug("[$clientId]:(gatewayClient) closePubConnection id: $id");
    pcController?.closePubConnection(id);
  }

  leave() async {
    Log.debug("[$clientId]:(gatewayClient) leave");
    try {
      state = LEAVING;
      final Map<String, dynamic> leaveReq = Message.generate("leave", [
        {"uid": uid}
      ]);
      await signal.request(leaveReq);
    } catch (e) {
      Log.debug("[$clientId]:(gatewayClient) send leave message failed: $e");
    } finally {
      Log.debug("[$clientId]:(gatewayClient) clear connection");
      chooseServer?.cancel();
      signal?.close();
      pcController?.clear();
      state = LEAVED;
    }
  }

  _onSiganlClose() {
    Log.debug("[$clientId]:(gatewayClient) _onSiganlClose");
    signal = null;
    pcController?.clear();
    if (state == CONNECTED || state == CONNECTING || state == JOINED) {
      state = DISCONNECTED;
    }
  }
}
