import 'dart:convert';
import 'dart:math';
import 'dart:async';
import '../utils/logger.dart';

import './message.dart';
import '../utils/compress.dart';
import './signal.dart';
import '../../agora_stream.dart';
import 'package:flutter_webrtc/webrtc.dart';

class PeerConnection {
  int p2pId;
  Signal signal;
  AgoraStream stream;
  RTCPeerConnection pc;
  List<Map<String, String>> stats;
  Map<String, dynamic> sessionInfo;
  Timer statsTimer;
  Timer reportTimer;

  static const Map<String, dynamic> _iceServers = {
    'iceServers': [
      {"url": "stun:webcs.agora.io:3478"}
    ]
  };

  static const Map<String, dynamic> _config = {
    'mandatory': {},
    'optional': [
      {'DtlsSrtpKeyAgreement': true},
    ],
  };

  static const Map<String, dynamic> _constraints = {
    'mandatory': {
      'OfferToReceiveAudio': true,
      'OfferToReceiveVideo': true,
    },
    'optional': [],
  };

  _getStats() async {
    var videoTracks = stream?.mediaStream?.getVideoTracks();
    var audioTracks = stream?.mediaStream?.getAudioTracks();
    var track = null;

    // 不管传入的track是什么类型，返回的结果都包含video和audio的stats
    if (videoTracks != null && videoTracks.length != 0) {
      track = videoTracks[0];
    }

    if (track == null && audioTracks != null && audioTracks.length != 0) {
      track = audioTracks[0];
    }

    if (track == null) {
      return;
    }

    var videoTrack = videoTracks[0];
    var stats = await pc.getStats(videoTrack);
    List<Map<String, String>> fStats = [];
    stats.forEach((stat) {
      if (stat.type == 'ssrc' || stat.type == 'VideoBwe') {
        Map<String, String> s = {
          "id": stat.id.toString(),
          "type": stat.type.toString(),
          "timestamp": stat.timestamp.toString(),
        };
        stat.values.forEach((key, value) => s[key] = value.toString());
        fStats.add(s);
      }
    });
    // debugPrint(fStats.toString(), wrapWidth: 1024);
    this.stats = fStats;
  }

  PeerConnection(this.signal, this.sessionInfo) {
    p2pId = Random().nextInt(DateTime.now().microsecond);
  }

  Future<PeerConnection> init() async {
    pc = await createPeerConnection(_iceServers, _config);

    // pc.onIceCandidate = (candidate) {
    //   print("candidate: $candidate");
    // };
    pc.onIceConnectionState = (state) {
      print("connection state: $state");
    };
    pc.onRenegotiationNeeded = () {
      print("on renegotiation needed");
    };
    pc.onIceGatheringState = (state) {
      print("ice gathering state: $state");
    };
    return this;
  }

  Future<AgoraStream> publish(AgoraStream stream) async {
    if (stream == null || stream.mediaStream == null) {
      throw "Invalid stream, stream or stream.mediaStream should not be null";
    }
    if (signal == null || signal.isConnected == false) {
      throw "Can't publish stream when signal is disconneted";
    }

    pc.addStream(stream?.mediaStream);
    RTCSessionDescription sdp1 = await pc.createOffer(_constraints);
    pc.setLocalDescription(sdp1);
    await _waitIceCandidate();
    RTCSessionDescription sdp = await pc.createOffer(_constraints);
    pc.setLocalDescription(sdp);

    final String sdpMsg = jsonEncode({
      "sdp": sdp.sdp,
      "messageType": "OFFER",
      "tiebreaker": Random().nextInt(429496723),
    });

    final Map<String, dynamic> p2pInfo = {
      "sdp": sdpMsg,
      "p2p_id": p2pId,
    };

    final Map<String, dynamic> publishReq =
        Message.generate("publish", [sessionInfo, p2pInfo, stream.streamInfo]);
    final Map<String, dynamic> answerMsg = await signal.request(publishReq);

    final msg = answerMsg["_message"];
    if (msg["error_code"] != null) {
      throw "${msg["error_code"]}: ${msg["error_str"]}";
    }

    var answerSDP = jsonDecode(msg["sdp"]);
    pc.setRemoteDescription(
        new RTCSessionDescription(answerSDP["sdp"], "answer"));

    await signal.gwEvent.firstWhere(
        (event) => event["_type"] == "on_publish_stream",
        orElse: () => {} // avoid print "Bad state: No element"
        );
    this.stream = stream;
    this.stream.pc = this;

    // get stats per sec
    statsTimer = Timer.periodic(new Duration(seconds: 1), (t) {
      _getStats();
    });

    // report media stats per 3 sec
    reportTimer = Timer.periodic(new Duration(seconds: 3), (t) {
      stats.forEach((stat) {
        var compressedStats = {};
        stat.forEach(
            (key, value) => compressedStats[Compress.encode(key)] = value);
        var publishStatsReq = Message.generate("publishStats", [
          {"stats": compressedStats}
        ]);
        signal.send(publishStatsReq);
      });
    });

    return stream;
  }

  Future<AgoraStream> unpublish(AgoraStream stream) async {
    if (stream == null || stream.mediaStream == null) {
      throw "Invalid stream, stream or stream.mediaStream should not be null";
    }
    if (signal == null || signal.isConnected == false) {
      throw "Can't unpublish stream when signal is disconneted";
    }

    final unpublishReq = Message.generate("unpublish", [stream.streamInfo]);
    try {
      await signal.request(unpublishReq);  
    } catch (e) {
      Log.error("[${stream?.streamId}] Send unpublish request failed, error: ${e.toString()}");
    }
    statsTimer.cancel();
    reportTimer.cancel();
    pc.removeStream(stream.mediaStream);
    try {
      await pc.close();  
    } catch (e) {
      Log.error("[${stream?.streamId}] Close publish peerConnection failed, error: ${e.toString()}");
    }
    stream.pc = null;
    return stream;
  }

  Future<AgoraStream> subscribe(AgoraStream stream) async {
    if (stream == null) {
      throw "Invalid stream, stream should not be null";
    }
    if (signal == null || signal.isConnected == false) {
      throw "Can't subscribe stream when signal is disconneted";
    }

    RTCSessionDescription sdp1 = await pc.createOffer(_constraints);
    pc.setLocalDescription(sdp1);
    await _waitIceCandidate();
    RTCSessionDescription sdp = await pc.createOffer(_constraints);
    pc.setLocalDescription(sdp);

    final String sdpMsg = jsonEncode({
      "sdp": sdp.sdp,
      "messageType": "OFFER",
      "tiebreaker": Random().nextInt(429496723),
    });

    final Map<String, dynamic> p2pInfo = {
      "sdp": sdpMsg,
      "p2p_id": p2pId,
    };

    final Map<String, dynamic> subscribeReq = Message.generate(
        "subscribe", [sessionInfo, p2pInfo, stream.streamInfo]);
    final Map<String, dynamic> answerMsg = await signal.request(subscribeReq);

    final msg = answerMsg["_message"];
    if (msg["error_code"] != null) {
      throw "${msg["error_code"]}: ${msg["error_str"]}";
    }

    var answerSDP = jsonDecode(msg["sdp"]);
    pc.setRemoteDescription(
        new RTCSessionDescription(answerSDP["sdp"], "answer"));

    MediaStream mediaStream = await _handleRemoteStream();
    stream.mediaStream = mediaStream;
    stream.pc = this;
    this.stream = stream;

    // get stats per sec
    this.statsTimer = Timer.periodic(new Duration(seconds: 1), (t) {
      _getStats();
    });

    // report media stats per 3 sec
    reportTimer = Timer.periodic(new Duration(seconds: 3), (t) {
      stats.forEach((stat) {
        var compressedStats = {};
        stat.forEach(
            (key, value) => compressedStats[Compress.encode(key)] = value);

        var subscribeStatsReq = Message.generate("subscribeStats", [
          {"stats": compressedStats},
          stream.streamInfo
        ]);
        signal.send(subscribeStatsReq);
      });
    });
    return stream;
  }

  Future<AgoraStream> unsubscribe(AgoraStream stream) async {
    if (stream == null || stream.mediaStream == null) {
      throw "Invalid stream, stream or stream.mediaStream should not be null";
    }
    if (signal == null || signal.isConnected == false) {
      throw "Can't unsubscribe stream when signal is disconneted";
    }

    final unsubscribeReq = Message.generate("unsubscribe", [stream.streamInfo]);
    try {
      await signal.request(unsubscribeReq);  
    } catch (e) {
      Log.error("[${stream?.streamId}] Send unsubscribe request failed, error: ${e.toString()}");
    }
    statsTimer.cancel();
    reportTimer.cancel();
    // 看flutter-webrtc的实现，removeStream只用于本地流，如果在这里调用会报错
    // await pc.removeStream(stream.mediaStream);
    try {
      await pc.close();  
    } catch (e) {
      Log.error("[${stream?.streamId}] Close subscribe peerConnection failed, error: ${e.toString()}");
    }
    stream.mediaStream = null;
    stream.pc = null;
    return stream;
  }

  Future<MediaStream> _handleRemoteStream() {
    Completer c = Completer<MediaStream>();
    Timer timeout = new Timer(Duration(seconds: 5), () {
      c.completeError("handle remote stream time out");
    });
    pc.onAddStream = (MediaStream stream) {
      pc.onAddStream = null;
      timeout.cancel();
      c.complete(stream);
    };
    return c.future;
  }

  Future<RTCIceCandidate> _waitIceCandidate() {
    Completer c = Completer<RTCIceCandidate>();
    Timer timeout = new Timer(Duration(seconds: 5), () {
      c.completeError("wait ice candidate time out");
    });
    pc.onIceCandidate = (RTCIceCandidate candidate) {
      pc.onIceCandidate = null;

      print(candidate.toMap());

      timeout.cancel();
      c.complete(candidate);
    };
    return c.future;
  }

  Future<void> close() async {
    statsTimer?.cancel();
    reportTimer?.cancel();
    // 看flutter-webrtc的实现，removeStream只用于本地流，如果在这里调用会报错
    if (stream != null && stream.mediaStream != null && stream.isLocal == true) {
      pc?.removeStream(stream.mediaStream);
    }
    await pc?.close();
    pc = null;
    signal = null;
    stream = null;
  }
}
