import 'dart:io';
import 'dart:core';
import 'dart:async';
import 'dart:convert';

import '../utils/logger.dart';

class Signal {
  WebSocket ws;
  String url;
  String clientId;
  Function onClose;
  StreamSubscription _messaageSubscription;

  int _requestId = 0;
  int _heartBeatFailedCount = 0;
  bool _isConnected = false;
  Stream _broadcastStream;
  Timer _heartBeatCheck;

  get state {
    return ws.readyState;
  }

  get isConnected {
    return _isConnected;
  }

  get gwEvent {
    return _broadcastStream
        ?.where((frame) => frame["_id"] == null && frame["_type"] != null);
  }

  get event {
    return _broadcastStream?.where((frame) => frame["_event"] != null);
  }

  static Future<Signal> connect(String url, clientId) {
    return new Signal().init(url, clientId);
  }

  Future<Signal> init(String url, clientId) async {
    this.clientId = clientId;
    try {
      url = url;
      ws = await WebSocket.connect("wss://$url");
      _isConnected = true;
      _broadcastStream =
          ws?.asBroadcastStream()?.map((data) => jsonDecode(data)); 
      ws.done.then((_) => _onClose());
      _startHeartBeatCheck();
    } catch (e) {
      return Future.error(e);
    }
    return Future.value(this);
  }

  void _startHeartBeatCheck() {
    _heartBeatCheck = Timer.periodic(Duration(seconds: 2), (t) {
      request({"_type": "ping"}).catchError((error) {
        print(error);
        if (++_heartBeatFailedCount >= 3) {
          // broke connection
          _heartBeatCheck.cancel();
          ws.close();
        }
      });
    });
  }

  void _add(String data) {
    // print("--- send --- ${(data)}");
    ws?.add(data);
  }

  void send(data) {
    _add(jsonEncode(data));
  }

  void _clear() {
    Log.debug("[$clientId]: clear signal state");
    _isConnected = false;
    _heartBeatCheck?.cancel();
    _messaageSubscription?.cancel();
  }

  void _onClose() {
    Log.debug("[$clientId]: ws onClose");
    _clear();
    if(onClose != null) {
      onClose();
    }
  }

  void close() {
    Log.debug("[$clientId]: close signal ");
    _clear();
    ws?.close();
  }

  Future request(data) {
    var reqId = "_request_${_requestId++}";
    data["_id"] = reqId;
    _add(jsonEncode(data));
    return _broadcastStream
        ?.firstWhere((response) => response["_id"] == reqId,
            orElse: () => {}) // avoid print "Bad state: No element"
        ?.timeout(Duration(seconds: 5));
  }

  StreamSubscription listen(Function onData) {
    return _broadcastStream?.listen(onData);
  }
}
